from dataclasses import is_dataclass
from typing import Union, List, Type, Optional, cast
from google.cloud.bigquery import SchemaField, TimePartitioning
from bq_schema.dataclass_converter import dataclass_to_schema


class BigqueryTable:
    @property
    def name(self) -> str:
        raise NotImplementedError

    @property
    def schema(self) -> Union[List[SchemaField], Type]:
        raise NotImplementedError

    @property
    def project(self) -> Optional[str]:
        return None

    @property
    def dataset(self) -> Optional[str]:
        return None

    @property
    def version(self) -> Optional[str]:
        return None

    @property
    def time_partitioning(self) -> Optional[TimePartitioning]:
        return None

    def full_table_name(self) -> str:
        if self.version:
            return f"{self.name}_v{self.version}"

        return self.name

    def get_schema_fields(self) -> List[SchemaField]:
        if is_dataclass(self.schema):
            return dataclass_to_schema(cast(Type, self.schema))

        return cast(List[SchemaField], self.schema)



from dataclasses import asdict
from typing import TypeVar, Generic, Type
from dacite import from_dict, Config

from google.cloud.bigquery.table import Row


T = TypeVar("T")  # pylint: disable=invalid-name


class RowTransformer(Generic[T]):
    def __init__(self, schema: Type[T]):
        self._schema: Type[T] = schema

    def bq_row_to_dataclass_instance(self, bq_row: Row) -> T:
        return from_dict(self._schema, bq_row, config=Config(check_types=False))

    @staticmethod
    def dataclass_instance_to_bq_row(instance: T) -> dict:
        return asdict(instance)



from typing import Optional
import argparse
from argparse import Namespace
from google.cloud import bigquery
from google.api_core.exceptions import NotFound
from google.cloud.bigquery.table import Table

from bq_schema.migration.table_finder import find_tables
from bq_schema.migration.schema_diff import find_new_columns


def parse_args() -> Namespace:
    parser = argparse.ArgumentParser(
        description="Manage your bigquery table migrations."
    )
    parser.add_argument("--project", required=False, help="Target bigquery project.")
    parser.add_argument("--dataset", required=False, help="Target bigquery dataset")
    parser.add_argument(
        "--module-path", required=True, help="Module path to your bigquery models."
    )

    parser.add_argument(
        "--apply",
        default=False,
        action="store_true",
        help="If set to true, this script will apply all open migrations.",
    )

    return parser.parse_args()


def main(
    project: Optional[str], dataset: Optional[str], module_path: str, apply: bool
) -> None:
    client = bigquery.Client()
    for local_table in set(find_tables(module_path)):
        project = project or local_table.project
        assert project, "Project has not been set."
        dataset = dataset or local_table.dataset
        assert dataset, "Dataset has not been set."

        table_identifier = f"{project}.{dataset}.{local_table.full_table_name()}"
        print(f"Checking migrations for: {table_identifier}")

        try:
            remote_table = client.get_table(table_identifier)
        except NotFound:
            print("Table does not exist in bq: {table_identifier}")
            if apply:
                print("Creating table.")
                table = Table(
                    table_identifier,
                    schema=local_table.get_schema_fields(),
                )
                if local_table.time_partitioning:
                    table.time_partitioning = local_table.time_partitioning
                print(client.create_table(table))
        else:
            new_columns = list(
                find_new_columns(local_table.get_schema_fields(), remote_table.schema)
            )
            if new_columns:
                print(f"Found new columns: {new_columns}")
                if apply:
                    print("Applying changes")
                    remote_table.schema = local_table.get_schema_fields()
                    print(client.update_table(remote_table, ["schema"]))


def cli() -> None:
    args = parse_args()
    main(args.project, args.dataset, args.module_path, args.apply)


if __name__ == "__main__":
    cli()



from google.cloud.bigquery import SchemaField
from bq_schema.migration.schema_diff import find_new_columns


def test_find_new_columns():
    local_schema = [SchemaField("a", "INTEGER")]
    assert list(find_new_columns(local_schema, [])) == local_schema


def test_find_new_columns_missing_record():
    local_schema = [SchemaField("a", "RECORD")]
    assert list(find_new_columns(local_schema, [])) == local_schema


def test_find_new_columns_missing_column_in_record():
    local_schema = [SchemaField("a", "RECORD", fields=[SchemaField("b", "INTEGER")])]
    remote_schema = [SchemaField("a", "RECORD", fields=[])]
    assert list(find_new_columns(local_schema, remote_schema)) == [
        SchemaField("b", "INTEGER", "NULLABLE", None, ())
    ]



from typing import List, Iterator
from google.cloud import bigquery


def find_new_columns(
    local_schema: List[bigquery.SchemaField],
    remote_schema: List[bigquery.SchemaField],
) -> Iterator[bigquery.SchemaField]:
    remote_columns = {column.name: column for column in remote_schema}

    for column in local_schema:
        if column.name not in remote_columns:
            yield column
            continue

        if column.field_type == "RECORD":
            yield from find_new_columns(
                column.fields, remote_columns[column.name].fields
            )
